# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from pygerber.gerberx3.api.v2 import GerberFile, ParsedFile
from pygerber.gerberx3.parser2.commands2.region2 import Region2
from pygerber.gerberx3.parser2.commands2.line2 import Line2
from pygerber.gerberx3.parser2.commands2.flash2 import Flash2
from pygerber.gerberx3.parser2.commands2.arc2 import Arc2, CCArc2
from pygerber.gerberx3.parser2.apertures2.polygon2 import Polygon2
from pygerber.gerberx3.parser2.apertures2.circle2 import Circle2
from pygerber.gerberx3.parser2.apertures2.rectangle2 import Rectangle2

from math import cos, sin, pi
import numpy as np
from itertools import groupby
import math

from ..._emerge import geo
from ..._emerge.cs import CoordinateSystem, GCS
from ..._emerge.geometry import GeoSurface, GeoPolygon
from ..._emerge.mth.loopsplit import Loop

from loguru import logger

N_CIRC_MIN = 6
N_CIRC_MAX = 12

def _calc_via_segs(diameter: float, nsegments: int, edge_size: float | None = None) -> int:
    if edge_size is not None:
        N = int(np.ceil(np.pi*diameter/edge_size))
    else:
        N = nsegments
    return max(min(N,N_CIRC_MAX),N_CIRC_MIN)
    
def zero_runs_indices_gb(xs: list[int], min_len=3):
    """Returns all index sequences of the input list of subsequent ids that form a segmen of at least min_len zeros.

    Args:
        xs (_type_): _description_
        min_len (int, optional): _description_. Defaults to 3.

    Returns:
        _type_: _description_
    """
    runs = []
    idx = 0
    for val, grp in groupby(xs):
        length = sum(1 for _ in grp)
        if val == 0 and length >= min_len:
            runs.append((idx, idx + length - 1))  # inclusive end
        idx += length
    return runs


############################################################
#                    POLYGON FUNCTIONS                    #
############################################################

def semi_circ(p: tuple[float, float], u: tuple[float, float], w: float, N: int, max_seg: float | None) -> list[tuple[float, float]]:
    N = _calc_via_segs(w, N, max_seg)
    x2, y2 = p
    ux, uy = u
    urx, ury = uy, -ux
    ths = np.linspace(0,np.pi,N)[1:-1]
    pes = []
    ex = w*np.sin(np.pi/(2*N))*ux 
    ey = w*np.sin(np.pi/(2*N))*uy
    for th in ths:
        c = (w*urx + w*ury*1j)*np.exp(1j*th)
        px = x2 + c.real
        py = y2 + c.imag
        pes.append((px+ex, py+ey))
    return pes

def parse_region(cmd: Region2) -> list[tuple[float, float]]:
    ring = []
    for seg in cmd.command_buffer:   # nested boundary commands
        if isinstance(seg, Line2):
            # points are Vector2D with unit helpers; use millimeters
            if not ring:
                ring.append((float(seg.start_point.x.as_millimeters()),
                            float(seg.start_point.y.as_millimeters())))
            ring.append((float(seg.end_point.x.as_millimeters()),
                        float(seg.end_point.y.as_millimeters())))
            continue
    return ring

def xy_to_c(points: list[tuple, tuple]) -> list[complex]:
    return [p[0]+1j*p[1] for p in points]

def c_to_xy(points: list[complex]) -> list[tuple[float, float]]:
    return [(p.real, p.imag) for p in points]

def dphi_cw(dr) -> tuple[float, float]:
    rx, ry = dr
    R = (rx**2 + ry**2)**0.5
    return (ry/R, -rx/R)

def dphi_ccw(dr) -> tuple[float, float]:
    rx, ry = dr
    R = (rx**2 + ry**2)**0.5
    return (-ry/R, rx/R)

def parse_arc(cmd: Arc2, ds: float, reverse: bool, Nseg: int, max_size: float | None) -> list[tuple[float, float]]:
    xc =  float(cmd.center_point.x.as_millimeters())
    yc =  float(cmd.center_point.y.as_millimeters())
    dx0 = float(cmd.get_relative_start_point().x.as_millimeters())
    dy0 = float(cmd.get_relative_start_point().y.as_millimeters())
    dx1 = float(cmd.get_relative_end_point().x.as_millimeters())
    dy1 = float(cmd.get_relative_end_point().y.as_millimeters())
    if reverse:
        dx0, dy0, dx1, dy1 = dx1, dy1, dx0, dy0
    w = float(cmd.aperture.diameter.as_millimeters())/2
    
    dang = np.angle((dx1+1j*dy1)/(dx0+1j*dy0)) #% (2*np.pi)
    if dang > 0:
        raise ValueError(f'Dang is positive {dang}')
    R = (dx0**2 + dy0**2)**0.5
    Nang = abs(int(np.ceil(1.5*dang*R/ds)))
    angs = np.linspace(0,dang,Nang)
    
    arm = dx0 + 1j*dy0
    cc = xc + 1j*yc
    psc1 = c_to_xy([cc+arm*np.exp(1j*th)*(R-w)/R for th in angs])
    psc2 = c_to_xy([cc+arm*np.exp(1j*th)*(R+w)/R for th in angs][::-1])
    
    end_cap = semi_circ((xc+dx1, yc+dy1), dphi_cw((dx1, dy1)), w, Nseg, max_size)
    begin_cap = semi_circ((xc+dx0, yc+dy0), dphi_ccw((dx0, dy0)), w, Nseg, max_size)
    
    points = psc1+end_cap+psc2+begin_cap
    return points

def parse_line(cmd: Line2, min_dist: float, ncirc:int , size_circ) -> list[tuple[float, float]]:
    diam = float(cmd.aperture.diameter.as_millimeters())
    x1 = float(cmd.start_point.x.as_millimeters())
    y1 = float(cmd.start_point.y.as_millimeters())
    x2 = float(cmd.end_point.x.as_millimeters())
    y2 = float(cmd.end_point.y.as_millimeters())
    
    L = ((x2-x1)**2 + (y2-y1)**2)**0.5
    
    if L < min_dist/10:
        diam = float(cmd.aperture.diameter.as_millimeters())
        x = float(cmd.start_point.x.as_millimeters())
        y = float(cmd.start_point.y.as_millimeters())
        R = diam/2 #+ diam/2*np.sin(np.pi/12)
        points = [(x+R*np.cos(th), y+R*np.sin(th)) for th in np.linspace(0,2*np.pi,_calc_via_segs(diam, ncirc, size_circ))]
        return points
    
    ux = (x2-x1)/L
    uy = (y2-y1)/L

    urx = uy
    ury = -ux
    w = diam/2
    
    p1 = (x1+w*urx, y1+w*ury)
    p2 = (x2+w*urx, y2+w*ury)
    p3 = (x2-w*urx, y2-w*ury)
    p4 = (x1-w*urx, y1-w*ury)
    
    pes = semi_circ((x2, y2), (ux, uy), w, ncirc, size_circ)
    pss = semi_circ((x1, y1), (-ux, -uy), w, ncirc, size_circ)
    
    points = [p1, p2] + pes + [p3, p4] + pss
    return points

def parse_flash_circ(cmd: Flash2, nsegments: int, segsize: float | None = None) -> list[tuple[float, float]]:
    diam = float(cmd.aperture.diameter.as_millimeters())
    x = float(cmd.flash_point.x.as_millimeters())
    y = float(cmd.flash_point.y.as_millimeters())
    
    N = _calc_via_segs(diam, nsegments, segsize)
    points = [(x+diam/2*np.cos(th), y+diam/2*np.sin(th)) for th in np.linspace(0,2*np.pi,N+1)]
    return points

def parse_flash_rect(cmd: Rectangle2) -> list[tuple[float, float]]:
    x = float(cmd.flash_point.x.as_millimeters())
    y = float(cmd.flash_point.y.as_millimeters())
    sizeX = float(cmd.aperture.x_size.as_millimeters())
    sizeY = float(cmd.aperture.y_size.as_millimeters())
    rotation = float(cmd.aperture.rotation)
    
    # Half sizes
    hx, hy = sizeX / 2.0, sizeY / 2.0

    # Local (unrotated) corners relative to center
    corners = [
        (-hx, -hy),
        ( hx, -hy),
        ( hx,  hy),
        (-hx,  hy),
    ]

    cos_r, sin_r = np.cos(rotation), np.sin(rotation)

    # Rotate + translate
    points = [
        (
            x + dx * cos_r - dy * sin_r,
            y + dx * sin_r + dy * cos_r
        )
        for dx, dy in corners
    ]

    return points

############################################################
#                          CLASSES                         #
############################################################


class ClosedPoly:
    
    def __init__(self, points: list[tuple[float, float]], clear: bool, simplify: bool):
        points = [(x*0.001, y*0.001) for x,y in points]
        if points[0] != points[-1]:
            points.append(points[0])
            
        points = self._ensure_clockwise(points)
        
        self.points: list[tuple[float, float]] = points
        self.clear: bool = clear
        x, y = zip(*self.points)
        self.xs: list[float] = x
        self.ys: list[float] = y
        self.do_simplify: bool = simplify
    
    @staticmethod
    def _ensure_clockwise(points: list[tuple[float, float]]):
        # Signed area (shoelace formula)
        area = 0.0
        for (x0, y0), (x1, y1) in zip(points, points[1:]):
            area += (x1 - x0) * (y1 + y0)

        # area > 0 → counter-clockwise → reverse
        if area > 0.0:
            return points[::-1]
        return points
    
    @property
    def dark(self) -> bool:
        return not self.clear
    
    @property
    def xy(self) -> tuple[list[float], list[float]]:
        return self.xs, self.ys
    
    def simplify(self, min_distance: float, min_incl_angle: float = 5, min_remove_angle: float = 10):
        """
        Simplify a closed polygon (last point is NOT repeated as first) by iteratively
        removing vertices according to the following rules:

        For each point i:
            dist_till_next = distance(i, i+1)
            dist_till_prev = distance(i, i-1)
            angle_next_prev = angle between vectors (prev -> i) and (i -> next)
                            in degrees, 0° = straight, 180° = U-turn

        A point is removed iff:
            |angle_next_prev| < min_incl_angle
        OR
            dist_till_next < min_distance AND
            dist_till_prev < min_distance AND
            |angle_next_prev| < min_remove_angle

        After any removal, the process restarts until no more points are removed.
        """

        # Optional guard: do nothing for very small polygons
        if len(self.points) <= 6:
            return

        # Work on local copies so we can pop safely
        xs = list(self.xs)
        ys = list(self.ys)
        n = len(xs)

        # Need at least a triangle
        if n < 3:
            return

        changed = True
        while changed and n > 6:
            changed = False

            for i in range(n):
                ip = (i - 1) % n        # previous index
                inext = (i + 1) % n     # next index

                x_i, y_i = xs[i], ys[i]
                x_p, y_p = xs[ip], ys[ip]
                x_n, y_n = xs[inext], ys[inext]

                # Distances to previous and next
                dx_prev = x_i - x_p
                dy_prev = y_i - y_p
                dx_next = x_n - x_i
                dy_next = y_n - y_i

                dist_prev = math.hypot(dx_prev, dy_prev)
                dist_next = math.hypot(dx_next, dy_next)

                # If either adjacent segment is degenerate, skip angle computation
                if dist_prev == 0.0 or dist_next == 0.0:
                    continue

                # Angle between (prev -> current) and (current -> next)
                dot = dx_prev * dx_next + dy_prev * dy_next
                denom = dist_prev * dist_next
                cos_angle = dot / denom

                # Clamp for numerical safety
                if cos_angle > 1.0:
                    cos_angle = 1.0
                elif cos_angle < -1.0:
                    cos_angle = -1.0

                angle = math.degrees(math.acos(cos_angle))  # in [0, 180]

                remove = False
                if abs(angle) < min_incl_angle:
                    remove = True
                elif (
                    dist_next < min_distance
                    and dist_prev < min_distance
                    and abs(angle) < min_remove_angle
                ):
                    remove = True

                # Do not reduce below a triangle
                if remove and n - 1 >= 6:
                    xs.pop(i)
                    ys.pop(i)
                    n -= 1
                    changed = True
                    break  # restart outer while-loop with updated polygon

        # Write back to object
        self.xs = xs
        self.ys = ys
        self.points = list(zip(xs, ys))



############################################################
#                       GERBER CLASS                      #
############################################################

class GerberLayer:
    
    def __init__(self, 
                 filename: str, 
                 res_mm: float,
                 n_circ_segments: int = 8,
                 seg_size_mm: float | None = None,
                 ignore_via_pads: bool = True,
                 cs: CoordinateSystem = GCS,
                 ):
        
        self.fname: str = filename
        self.pf: ParsedFile = GerberFile.from_file(filename).parse()
        self.buffer = self.pf._command_buffer
        self.polies: list[ClosedPoly] = []
        self.res_mm: float = res_mm
        
        self._nseg: int = n_circ_segments
        self._seg_size_mm: float = seg_size_mm
        
        self.xs = []
        self.ys = []
        
        self.cs: CoordinateSystem = cs
        self.parse_file()
        self.simplify()
    
    def bounds(self, margin: float | tuple[float, float, float, float] = 0.0) -> tuple[float, float, float, float]:
        if isinstance(margin, (float, int)):
            return min(self.xs)-margin, min(self.ys)-margin, max(self.xs)+margin, max(self.ys)+margin
        else:
            return min(self.xs)-margin[0], min(self.ys)-margin[1], max(self.xs)+margin[2], max(self.ys)+margin[3]
        
    def _add_poly(self, points: list[tuple], clear: bool, simplify: bool) -> None:
        if points[-1]==points[0]:
            points = points[:-1]
        poly = ClosedPoly(points, clear, simplify)
        self.polies.append(poly)
        self.xs.extend(poly.xs)
        self.ys.extend(poly.ys)
        
    
    def parse_file(self):
        for cmd in self.buffer:
            
            if 'dark' in str(cmd.transform.polarity).lower():
                clear = False
            else:
                clear = True
            
            if isinstance(cmd, Region2):
                poly = parse_region(cmd)
                self._add_poly(poly, clear, True)
                continue
            
            if isinstance(cmd, CCArc2):
                points = parse_arc(cmd, self.res_mm, True, self._nseg, self._seg_size_mm)
                self._add_poly(points, clear, True)
                continue
            
            if isinstance(cmd, Arc2):
                points = parse_arc(cmd, self.res_mm, True, self._nseg, self._seg_size_mm)
                self._add_poly(points, clear, True)
                continue
            
            if isinstance(cmd, Line2):
                points = parse_line(cmd, self.res_mm, self._nseg, self._seg_size_mm)
                self._add_poly(points, clear, True)
                continue
            
            if not isinstance(cmd, Flash2):
                print('Not a Flash2, Uprocessed command:', cmd)
                continue
            
            if isinstance(cmd.aperture, Polygon2):
                ap = cmd.aperture
                # Field names are available on the dataclass; inspect if unsure:
                # print(ap.model_dump())
                n = int(ap.vertices)             # number of sides
                R = float(ap.diameter.as_millimeters()) / 2.0
                # Flash placement lives in the command transform:
                tr = cmd.transform               # translation + rotation
                cx = float(tr.translation.x.as_millimeters())
                cy = float(tr.translation.y.as_millimeters())
                # Rotation is in degrees in PyGerber’s apertures; convert to radians if needed
                rot = float(ap.rotation) * pi / 180.0
                verts = [(cx + R * cos(2*pi*i/n + rot), cy + R * sin(2*pi*i/n + rot))
                        for i in range(n)]
                self._add_poly(verts, clear, True)
                continue
            
            if isinstance(cmd.aperture, Circle2):
                points = parse_flash_circ(cmd, self._nseg, self._seg_size_mm)
                self._add_poly(points, clear, True)
                continue
            
            if isinstance(cmd.aperture, Rectangle2):
                points = parse_flash_rect(cmd)
                self._add_poly(points, clear, True)
                continue
            logger.error(f'Unparsable command found! {cmd}. Please contact the EMerge developers with this Gerber file to get this command supported.')
     
    def simplify(self):
        for poly in self.polies:
            poly.simplify(self.res_mm)
    
    def flatten(self, z: float = 0.0) -> GeoSurface:
        """Merge a GerberLayer into a GeoSurface

        Raises:
            Exception: _description_

        Returns:
            GeoSurface: _description_
        """
        poly = None
        
        xall = []
        yall = []
        import gmsh

        for isdark, polygons in groupby(self.polies, key=lambda x: x.dark):
            if poly is None and not isdark:
                raise Exception('First polygon is a clear:(')
            surfs = []
            for pol in polygons:
                x, y = pol.xy
                loop = Loop(np.array(x),np.array(y))
                add, remove = loop.split()
                LA = []
                LR = []
                for ax, ay in add:
                    LA.append(geo.XYPolygon(ax, ay, resolution=1e-7).geo(self.cs.displace(0,0,z)))
                    xall.extend(ax)
                    yall.extend(ay)
                for ax, ay in remove:
                    LR.append(geo.XYPolygon(ax, ay, resolution=1e-7).geo(self.cs.displace(0,0,z)))
                
                if len(LA)==1 and len(LR)==0:
                    surfs.append(LA[0])
                    continue
                
                if len(LR)==0:
                    surf = geo.unite(*LA)
                    surfs.append(surf)
                    continue
                
                surf = geo.subtract(geo.unite(*LA), geo.unite(*LR))
                surfs.append(surf)
           
            self.xs = xall
            self.ys = yall
            
            if poly is None:
                poly = geo.unite(*surfs)
                continue
            
            if isdark:
                poly = geo.unite(poly, *surfs)
            else:
                poly = geo.subtract(poly, geo.unite(*surfs))
            
            gmsh.fltk.run()
        return poly
            


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import gmsh # type: ignore
from emsutil import Material
from emsutil.lib import AIR
from .selection import FaceSelection, DomainSelection, EdgeSelection, PointSelection, Selection, SelectionError
from loguru import logger
from typing import Literal, Any, Iterable, Callable
import numpy as np


def _map_tags(tags: list[int], mapping: dict[int, list[int]]):
    new_tags = []
    for tag in tags:
        new_tags.extend(mapping.get(tag, [tag,]))
    return new_tags

def _bbcenter(x1, y1, z1, x2, y2, z2):
    return np.array([(x1+x2)/2, (y1+y2)/2, (z1+z2)/2])


FaceNames = Literal['back','front','left','right','top','bottom']

class _KEY_GENERATOR:

    def __init__(self):
        self.start = -1
    
    def new(self) -> int:
        self.start += 1
        return self.start

class _GeometryManager:

    def __init__(self):
        self.geometry_list: dict[str, dict[str, GeoObject]] = dict()
        self.active: str = ''
        self.geometry_names: dict[str, set[str]] = dict()

    def get_surfaces(self) -> list[GeoSurface]:
        return [geo for geo in self.all_geometries() if geo.dim==2]
    
    def all_geometries(self, model: str | None = None) -> list[GeoObject]:
        if model is None:
            model = self.active
        return [geo for geo in self.geometry_list[model].values() if geo._exists]

    def set_geometries(self, geos: list[GeoObject], model: str | None = None):
        if model is None:
            model = self.active
        self.geometry_list[model] = {geo.name: geo for geo in geos}
        
    def all_names(self, model: str | None = None) -> set[str]:
        if model is None:
            model = self.active
        return self.geometry_names[model]
    
    def get_name(self, suggestion: str, model: str | None = None) -> str:
        names = self.all_names(model)
        if suggestion not in names:
            return suggestion
        for i in range(1_000_000):
            if f'{suggestion}_{i}' not in names:
                return f'{suggestion}_{i}'
        raise RuntimeError('Cannot generate a unique name.')
        
    def submit_geometry(self, geo: GeoObject, model: str | None = None) -> None:
        if model is None:
            model = self.active
        self.geometry_list[model][geo.name] = geo
        self.geometry_names[model].add(geo.name)

    def sign_in(self, modelname: str) -> None:
        self.geometry_list[modelname] = dict()
        self.geometry_names[modelname] = set()
        self.active = modelname

    def reset(self, modelname: str) -> None:
        self.sign_in(modelname)
    
    def lowest_priority(self) -> int:
        return min([geo._base_priority for geo in self.all_geometries()])
    
    def highest_priority(self) -> int:
        return max([geo._base_priority for geo in self.all_geometries()])
    
class _FacePointer:
    """The FacePointer class defines a face to be selectable as a
    face normal vector plus an origin. All faces of an object
    can be selected based on the projected distance to the defined
    selection plane of the center of mass of a face iff the normals
    also align with some tolerance.

    """
    def __init__(self, 
                 origin: np.ndarray, 
                 normal: np.ndarray):
        self.o = np.array(origin)
        self.n = np.array(normal)

    def find(self, dimtags: list[tuple[int,int]],
             origins: list[np.ndarray],
             normals: list[np.ndarray]) -> list[int]:
        tags = []
        for (d,t), o, n in zip(dimtags, origins, normals):
            normdist = np.abs((o-self.o) @ self.n)
            dotnorm = np.abs(n@self.n)
            if normdist < 1e-5 and dotnorm > 0.999:
                tags.append(t)
        return tags
    
    def rotate(self, c0, ax, angle):
        """
        Rotate self.o and self.n about axis `ax`, centered at `c0`, by `angle` radians.

        Parameters
        ----------
        c0 : np.ndarray
            The center of rotation, shape (3,).
        ax : np.ndarray
            The axis to rotate around, shape (3,). Need not be unit length.
        angle : float
            Rotation angle in radians.
        """
        angle = -angle
        # Ensure axis is a unit vector
        k = ax / np.linalg.norm(ax)

        # Precompute trig values
        cos_theta = np.cos(angle)
        sin_theta = np.sin(angle)

        def rodrigues(v: np.ndarray) -> np.ndarray:
            """
            Rotate vector v around axis k by angle using Rodrigues' formula.
            """
            # term1 = v * cosθ
            term1 = v * cos_theta
            # term2 = (k × v) * sinθ
            term2 = np.cross(k, v) * sin_theta
            # term3 = k * (k ⋅ v) * (1 - cosθ)
            term3 = k * (np.dot(k, v)) * (1 - cos_theta)
            return term1 + term2 + term3

        # Rotate the origin point about c0:
        rel_o = self.o - c0            # move to rotation-centre coordinates
        rot_o = rodrigues(rel_o)       # rotate
        self.o = rot_o + c0            # move back

        # Rotate the normal vector (pure direction, no translation)
        self.n = rodrigues(self.n)

    def translate(self, dx, dy, dz):
        self.o = self.o + np.array([dx, dy, dz])
    
    def mirror(self, c0: np.ndarray, pln: np.ndarray) -> None:
        """
        Reflect self.o and self.n across the plane passing through c0
        with normal pln.

        Parameters
        ----------
        c0 : np.ndarray
            A point on the mirror plane, shape (3,).
        pln : np.ndarray
            The normal of the mirror plane, shape (3,). Need not be unit length.
        """
        # Normalize the plane normal
        k = pln / np.linalg.norm(pln)
        

        # Reflect the origin point:
        # compute vector from plane point to self.o
        v_o = self.o - c0
        # signed distance along normal
        dist_o = np.dot(v_o, k)
        # reflection
        self.o = self.o - 2 * dist_o * k

        # Reflect the normal/direction vector:
        dist_n = np.dot(self.n, k)
        self.n = self.n - 2 * dist_n * k

    def affine_transform(self, M: np.ndarray):
        """
        Apply a 4×4 affine transformation matrix to both self.o and self.n.

        Parameters
        ----------
        M : np.ndarray
            The 4×4 affine transformation matrix.
            - When applied to a point, use homogeneous w=1.
            - When applied to a direction/vector, use homogeneous w=0.
        """
        # Validate shape
        if M.shape != (4, 4):
            raise ValueError(f"Expected M to be 4×4, got shape {M.shape}")

        # Transform origin point (homogeneous w=1)
        homo_o = np.empty(4)
        homo_o[:3] = self.o
        homo_o[3] = 1.0
        transformed_o = M @ homo_o
        self.o = transformed_o[:3]

        # Transform normal/direction vector (homogeneous w=0)
        homo_n = np.empty(4)
        homo_n[:3] = self.n
        homo_n[3] = 0.0
        transformed_n = M @ homo_n
        self.n = transformed_n[:3]
        # Optionally normalize self.n if you need to keep it unit-length:
        # self.n = self.n / np.linalg.norm(self.n)

    def copy(self) -> _FacePointer:
        return _FacePointer(self.o, self.n)

    def __eq__(self, other: _FacePointer) -> bool:
        return (np.linalg.norm(self.o - other.o) + np.linalg.norm(self.n - other.n)) < 1e-6

_GENERATOR = _KEY_GENERATOR()
_GEOMANAGER = _GeometryManager()

class GeoObject:
    """A generalization of any OpenCASCADE entity described by a dimension and a set of tags.
    """
    dim: int = -1
    _default_name: str = 'GeoObject'
    
    def __init__(self, tags: list[int] | None = None, name: str | None = None):
        if tags is None:
            tags = []
        self.old_tags: list[int] = []
        self.tags: list[int] = tags
        self.material: Material = AIR
        self.mesh_multiplier: float = 1.0
        self.max_meshsize: float = 1e9
        
        self._unset_constraints: bool = False
        self._embeddings: list[GeoObject] = []
        self._face_pointers: dict[str, _FacePointer] = dict()
        self._tools: dict[int, dict[str, _FacePointer]] = dict()
        self._hidden: bool = False
        self._key = _GENERATOR.new()
        self._aux_data: dict[str, Any] = dict()
        self._base_priority: int = 10.0
        self._sub_priority: int = 0

        self._exists: bool = True
        
        self.give_name(name)
        _GEOMANAGER.submit_geometry(self)
        self._fill_face_pointers()
        
    @property
    def _priority(self) -> float:
        """The Priority of the geometry material

        Returns:
            float: _description_
        """
        return self._base_priority + self._sub_priority / 2
    
    def _fill_face_pointers(self) -> None:
        """ Fills the list of all face pointers of this object
        """
        current = list(self._face_pointers.values())
        ctr = 0
        gmsh.model.occ.synchronize()
        for dim, tag in gmsh.model.get_boundary(self.dimtags, True, False):
            if dim != 2:
                continue
            o = gmsh.model.occ.get_center_of_mass(2, tag)
            n = gmsh.model.get_normal(tag, (0,0))
            fp = _FacePointer(o, n)
            if fp not in current:
                self._face_pointers[f'Face{ctr}'] = fp
                ctr += 1
            
    def _store(self, name: str, data: Any) -> None:
        """Store a property as auxilliary data under a given name

        Args:
            name (str): Name field
            data (Any): Data to store
        """
        self._aux_data[name] = data

    def _load(self, name: str) -> Any | None:
        """Load data with a given name. If it doesn't exist, it returns None

        Args:
            name (str): The property to retreive

        Returns:
            Any | None: The property
        """
        return self._aux_data.get(name, None)
    
    def give_name(self, name: str | None = None) -> GeoObject:
        """Assign a name to this object

        Args:
            name (str | None, optional): The name for the object. Defaults to None.
        """
        if name is None:
            name = self._default_name
        self.name: str = _GEOMANAGER.get_name(name)
        return self
    
    @property
    def color_rgb(self) -> tuple[float, float, float]:
        """The color of the object in RGB float tuple

        Returns:
            tuple[float, float, float]: The color
        """
        return self.material.color_rgb
    
    @property
    def opacity(self) -> float:
        """The opacity of the object

        Returns:
            float: The opacity
        """
        return self.material.opacity
    
    @property
    def _metal(self) -> bool:
        """If the material should be rendered as metal
        """
        return self.material._metal
    
    @property
    def selection(self) -> Selection:
        '''Returns a corresponding Face/Domain or Edge Selection object'''
        if self.dim==1:
            return EdgeSelection(self.tags)
        elif self.dim==2:
            return FaceSelection(self.tags)
        elif self.dim==3:
            return DomainSelection(self.tags)
        else:
            return Selection(self.tags)
    
    @staticmethod
    def merged(objects: list[GeoPoint | GeoEdge | GeoSurface | GeoVolume | GeoObject]) -> list[GeoPoint | GeoEdge | GeoSurface | GeoVolume | GeoObject] | GeoPoint | GeoEdge | GeoSurface | GeoVolume | GeoObject:
        """Create a GeoObject by merging an iterable of GeoObjects

        Args:
            objects (list[GeoPoint  |  GeoEdge  |  GeoSurface  |  GeoVolume  |  GeoObject]): A list of geo objects

        Returns:
            GeoPoint | GeoEdge | GeoSurface | GeoVolume | GeoObject: The resultant object
        """
        dim = objects[0].dim
        tags = []
        out: GeoObject | None = None
        for obj in objects:
            tags.extend(obj.tags)
            obj.deactivate()
        
        if dim==2:
            out = GeoSurface(tags)
        elif dim==3:
            out = GeoVolume(tags)
        else:
            out = GeoObject(tags)
            
        out.material = objects[0].material
        out.prio_set(objects[0]._priority)
        out.name = f'MergedGeometries{[obj.name for obj in objects]}'
        return out
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.name},{self.dim},{self.tags})'

    def _data(self, *labels) -> tuple[Any | None, ...]:
        return tuple([self._aux_data[lab] for lab in labels])
    
    def _replace_pointer(self, name: str, face_pointer: _FacePointer) -> None:
        """Will be used to replace face pointers so only one unique one exists.

        Args:
            name (str): _description_
            face_pointer (_FacePointer): _description_
        """
        for key, fp in self._face_pointers.items():
            if fp == face_pointer:
                self._face_pointers.pop(key)
                break
        self._face_pointers[name] = face_pointer

    def _add_face_pointer(self, 
                          name: str,
                          origin: np.ndarray | None = None,
                          normal: np.ndarray | None = None,
                          tag: int | None = None):
        """Adds a face identifier (face pointer) to this object

        Args:
            name (str): The name for the face
            origin (np.ndarray | None, optional): A point on the object. Defaults to None.
            normal (np.ndarray | None, optional): The normal of the face. Defaults to None.
            tag (int | None, optional): The tace tag used to extract the origin and normal. Defaults to None.

        Raises:
            ValueError: _description_
        """
        if tag is not None:
            origin = gmsh.model.occ.get_center_of_mass(2, tag)
            normal = gmsh.model.get_normal(tag, (0,0))
        
        if origin is None or normal is None:
            raise ValueError('Eitehr a tag or an origin + normal must be provided!')
        fp = _FacePointer(origin, normal)
        self._face_pointers[name] = fp
        #self._replace_pointer(name, fp) <-- Will be added in later versions
        
        
    def make_copy(self) -> GeoObject:
        """ Copies this object and returns a new object (also in GMSH)"""
        new_dimtags = gmsh.model.occ.copy(self.dimtags)
        new_obj = GeoObject.from_dimtags(new_dimtags)
        new_obj.material = self.material
        new_obj.mesh_multiplier = self.mesh_multiplier
        new_obj.max_meshsize = self.max_meshsize
        
        new_obj._unset_constraints = self._unset_constraints
        new_obj._embeddings = [emb.make_copy() for emb in self._embeddings]
        new_obj._face_pointers = {key: value.copy() for key,value in self._face_pointers.items()}
        new_obj._tools = {key: {key2: value2.copy() for key2, value2 in value.items()} for key,value in self._tools.items()}
        
        new_obj._aux_data = self._aux_data.copy()
        new_obj._base_priority = self._base_priority
        new_obj._exists = self._exists
        return new_obj

    def replace_tags(self, tagmap: dict[int, list[int]]):
        """Replaces the GMSH tags assigned to this objects

        Args:
            tagmap (dict[int, list[int]]): A map that shows which tag is mapped to which set of new tags.
        """
        self.old_tags = self.tags
        newtags = []
        for tag in self.tags:
            newtags.extend(tagmap.get(tag, [tag,]))
        self.tags = newtags
        logger.debug(f'{self} Replaced {self.old_tags} -> {self.tags}')
    
    def update_tags(self, tag_mapping: dict[int,dict]) -> GeoObject:
        ''' Update the tag definition of a GeoObject after fragementation.'''
        self.replace_tags(tag_mapping[self.dim])
        return self
    
    def _take_pointers(self, *others: GeoObject) -> GeoObject:
        for other in others:
            self._face_pointers.update(other._face_pointers)
            self._tools.update(other._tools)
        return self
    
    @property
    def _all_pointers(self) -> list[_FacePointer]:
        pointers = list(self._face_pointers.values())
        for dct in self._tools.values():
            pointers.extend(list(dct.values()))
        return pointers
    
    @property
    def _all_pointer_names(self) -> set[str]:
        keys = set(self._face_pointers.keys())
        for dct in self._tools.values():
            keys = keys.union(set(dct.keys()))
        return keys
    
    def _take_tools(self, *objects: GeoObject) -> GeoObject:
        for obj in objects:
            self._tools[obj._key] = obj._face_pointers
            self._tools.update(obj._tools)
        return self
    
    def _face_tags(self, name: FaceNames, tool: GeoObject | None = None) -> list[int]:
        names = self._all_pointer_names
        if name not in names:
            raise ValueError(f'The face {name} does not exist in {self}. Only {list(self._face_pointers.keys())}')
        
        gmsh.model.occ.synchronize()
        dimtags = gmsh.model.get_boundary(self.dimtags, True, False)
        
        normals = [gmsh.model.get_normal(t, [0,0]) for d,t, in dimtags]
        origins = [gmsh.model.occ.get_center_of_mass(d, t) for d,t in dimtags]
        
        if tool is not None:
            tags = self._tools[tool._key][name].find(dimtags, origins, normals)
        else:
            tags = self._face_pointers[name].find(dimtags, origins, normals)
        logger.info(f'Selected face {tags}.')
        return tags

    
    def set_material(self, material: Material) -> GeoObject:
        """Assign a material to this geometry object.

        Args:
            material (Material): The material to assign

        Returns:
            GeoObject: This same object
        """
        self.material = material
        self._prio_half_up()
        return self
    
    def prio_set(self, level: int) -> GeoObject:
        """Defines the material assignment priority level of this geometry.
        By default all objects have priority level 10. If you assign a lower number,
        in cases where multiple geometries occupy the same volume, the highest priority
        will be chosen.

        Args:
            level (int): The priority level

        Returns:
            GeoObject: The same object
        """
        self._base_priority = int(level)
        return self
    
    def _prio_half_up(self) -> GeoObject:
        """Adds one half to the priority

        Returns:
            GeoObject: _description_
        """
        
        self._sub_priority = 1
        
    def above(self, other: GeoObject) -> GeoObject:
        """Puts the priority of this object one higher than the other, then returns this object

        Args:
            other (GeoObject): The other object to put below this object

        Returns:
            GeoObject: This object
        """
        self._base_priority = other._base_priority + 1.0
        return self
    
    def below(self, other: GeoObject) -> GeoObject:
        """Puts the priority of this object one lower than the other, then returns this object

        Args:
            other (GeoObject): The other object to put above this object

        Returns:
            GeoObject: This object
        """
        self._base_priority = other._base_priority - 1
        return self
        
    def prio_up(self) -> GeoObject:
        """Increases the material selection priority by 1

        Returns:
            GeoObject: _description_
        """
        self._base_priority += 1
        return self
    
    def prio_down(self) -> GeoObject:
        """Decreases the material selection priority by 1

        Returns:
            GeoObject: _description_
        """
        self._base_priority -= 1
        return self

    def background(self) -> GeoObject:
        """Set the material selection priority to be on the background.

        Returns:
            GeoObject: _description_
        """
        self._base_priority = _GEOMANAGER.lowest_priority()-10
        return self

    def foreground(self) -> GeoObject:
        """Set the material selection priority to be on top.

        Returns:
            GeoObject: _description_
        """
        self._base_priority = _GEOMANAGER.highest_priority()+10
        return self
    
    def boundary(self, 
                 exclude: Iterable[FaceNames,...] | str | None = None, 
                 tags: list[int] | None = None,
                 tool: GeoObject | None = None) -> FaceSelection:
        """Returns the complete set of boundary faces.
        
        If implemented, it is possible to exclude a set of faces based on their name
        or a list of tags.
        
        Args:
            exclude: (Iterable[str], str, None): A single string or list/tuple of strings.
            tags: A list of face integers (if known)
            tool: The tool object to base the selection face names one.

        Returns:
            FaceSelection: The selected faces
        """
        # if not self._exists:
        #     raise SelectionError('Cannot select faces from an object that no longer exists.')
        
        if isinstance(exclude, str):
            exclude = (exclude,)
            
        if exclude is None:
            exclude = tuple()
            
        if tags is None:
            tags = []
        
        for name in exclude:
            tags.extend(self.face(name, tool=tool).tags)
        dimtags = gmsh.model.get_boundary(self.dimtags, True, False)
        selname = 'Boundary'
        if exclude:
            selname = selname + 'Except[' + ','.join(exclude) + ']'
        return FaceSelection([t for d,t in dimtags if t not in tags])._named(selname)
    
    def face(self, name: FaceNames = None, tool: GeoObject | None = None, no: FaceNames = None) -> FaceSelection:
        """Returns the FaceSelection for a given face name.
        
        The face name must be defined for the type of geometry.

        FaceNames include: front, back, left, right, top, bottom, disc
        
        Args:
            name (FaceNames): The name of the face to select.
            tool (GeoObject, None): Which object should be used as a source for the face selection.
            no (FaceNames): If everything BUT a face name should be selected, Equivalent to .boundary(exclude=name).

        Returns:
            FaceSelection: The selected face
        """
        if no is not None:
            return self.boundary(exclude=no)
        
        return FaceSelection(self._face_tags(name, tool))._named(name)

    def all_faces(self, tool: GeoObject | None = None) -> list[FaceSelection]:
        """Returns a list of all face selections of this object

        Returns:
            list[FaceSelection]: A list of all face selections
        """
        if tool is not None:
            return [self.face(name, tool=tool) for name in self._tools[tool._key].keys()]
        return [self.face(name) for name in self._face_pointers]
    
    def faces(self, *names: FaceNames, tool: GeoObject | None = None) -> FaceSelection:
        """Returns the FaceSelection for a given face names.
        
        The face name must be defined for the type of geometry.

        Args:
            name (FaceNames): The name of the face to select.
            tool (GeoObject, None): The tool object to use as source of the selection.

        Returns:
            FaceSelection: The selected face
        """
        tags = []
        for name in names:
            tags.extend(self._face_tags(name, tool))
        return FaceSelection(tags)._named('Faces[' + ','.join(names) + ']')
    
    def hide(self) -> GeoObject:
        """Hides the object from views

        Returns:
            GeoObject: _description_
        """
        self._hidden = True
        return self
    
    def unhide(self) -> GeoObject:
        """Unhides the object from views

        Returns:
            GeoObject: _description_
        """
        self._hidden = False
        return self
    
    @property
    def dimtags(self) -> list[tuple[int, int]]:
        return [(self.dim, tag) for tag in self.tags]
    
    @property
    def embeddings(self) -> list[tuple[int,int]]:
        return []

    @staticmethod
    def from_dimtags(dimtags: list[tuple[int,int]]) -> GeoVolume | GeoSurface | GeoObject:
        dim = dimtags[0][0]
        tags = [t for d,t in dimtags]
        if dim==0:
            return GeoPoint(tags)
        elif dim==1:
            return GeoEdge(tags)
        if dim==2:
            return GeoSurface(tags)
        if dim==3:
            return GeoVolume(tags)
        return GeoObject(tags)
    
    def remove(self) -> None:
        """Delete the geometry from the simulation model
        """
        self._exists = False
        gmsh.model.occ.remove(self.dimtags, True)
    
    def deactivate(self) -> None:
        """Make the geometry not considered as a valid EMerge geometry
        """
        self._exists = False
        
    def extract(self, tags: int | list[int]) -> GeoObject:
        """Returns a new GeoObject of the same dimensional type that isolates a set of given tags

        Args:
            tags (list[int]): A list of GMSH tags

        Returns:
            GeoObject: _description_
        """
        if isinstance(tags, int):
            tags = [tags,]
        self.tags = [t for t in self.tags if t not in tags]
        dts = [(self.dim, t) for t in tags]
        return GeoObject.from_dimtags(dts)
    
class GeoVolume(GeoObject):
    '''GeoVolume is an interface to the GMSH CAD kernel. It does not represent EMerge
    specific geometry data.'''
    dim = 3
    _default_name: str = 'GeoVolume'
    
    def __init__(self, tag: int | Iterable[int], name: str | None = None):
        super().__init__(name=name)
        
        self.tags: list[int] = []
        
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]

        self._fill_face_pointers()
        self._autoname()
        
    @property
    def selection(self) -> DomainSelection:
        return DomainSelection(self.tags)
    
    def _auto_face_tag(self) -> GeoVolume:
        self._face_pointers = dict()
        self._tools = dict()
        logger.trace('Automatically assigning face pointers.')
        ctr = 0
        for tag in self.tags:
            loops = gmsh.model.occ.getSurfaceLoops(tag)
            for loopcluster in loops[1]:
                for st in loopcluster:
                    self._add_face_pointer(f'Face{ctr}', tag=st)
                    ctr += 1
        return self
    
    def exterior_faces(self, base_object: GeoObject) -> FaceSelection:
        """Select the exterior faces of an object based on the face
        pointers of an original base object. For example
        
        Example:
        >>> cheese = em.geo.Box(...)
        >>> hole = em.geo.Box(...) 
        >>> holed_cheese = em.geo.subtract(cheese, hole)
        >>> holed_cheese.exterior_faces(cheese)

        Args:
            base_object (GeoObject): The object to base the selection on.

        Returns:
            FaceSelection: The resultant face selection
        """
        dimtags = gmsh.model.get_boundary(self.dimtags, True, False)
        
        normals = [gmsh.model.get_normal(t, [0,0]) for d,t, in dimtags]
        origins = [gmsh.model.occ.get_center_of_mass(d, t) for d,t in dimtags]
        
        tool_tags = []
        for key, tool in self._tools.items():
            if key == base_object._key:
                continue
            for name in tool.keys():
                tags = tool[name].find(dimtags, origins, normals)
                tool_tags.extend(tags)
        tags = [dt[1] for dt in dimtags if dt[1] not in tool_tags]
        return FaceSelection(tags)
    
    def _find_fp(self, condition: Callable):
        for key, fp in self._face_pointers.items():
            if condition(fp):
                return key, fp
        return None, None
    
    def _rename_fp(self, new_name: str, condition: Callable):
        name, fp = self._find_fp(condition)
        if fp is not None:
            self._face_pointers.pop(name)
            self._face_pointers[new_name] = fp
        
    def _autoname(self) -> None:
        if len(self._face_pointers)==0:
            return
        
        xs = []
        ys = []
        zs = []
        for fp in self._face_pointers.values():
            xs.append(fp.o[0])
            ys.append(fp.o[1])
            zs.append(fp.o[2])
        minx = min(xs)
        maxx = max(xs)
        miny = min(ys)
        maxy = max(ys)
        minz = min(zs)
        maxz = max(zs)
        
        self._rename_fp('-x', lambda fp: (fp.o[0]==minx) and np.abs(np.dot(fp.n, np.array([1,0,0])))>0.999)
        self._rename_fp('+x', lambda fp: (fp.o[0]==maxx) and np.abs(np.dot(fp.n, np.array([1,0,0])))>0.999)
        self._rename_fp('-y', lambda fp: (fp.o[1]==miny) and np.abs(np.dot(fp.n, np.array([0,1,0])))>0.999)
        self._rename_fp('+y', lambda fp: (fp.o[1]==maxy) and np.abs(np.dot(fp.n, np.array([0,1,0])))>0.999)
        self._rename_fp('-z', lambda fp: (fp.o[2]==minz) and np.abs(np.dot(fp.n, np.array([0,0,1])))>0.999)
        self._rename_fp('+z', lambda fp: (fp.o[2]==maxz) and np.abs(np.dot(fp.n, np.array([0,0,1])))>0.999)
        
class GeoPoint(GeoObject):
    dim = 0
    _default_name: str = 'GeoPoint'
    
    @property
    def selection(self) -> PointSelection:
        return PointSelection(self.tags)
    
    def __init__(self, tag: int | list[int], name: str | None = None):
        super().__init__(name=name)

        self.tags: list[int] = []
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]

class GeoEdge(GeoObject):
    dim = 1
    _default_name: str = 'GeoEdge'
    
    @property
    def selection(self) -> EdgeSelection:
        return EdgeSelection(self.tags)
    
    def __init__(self, tag: int | list[int], name: str | None = None):
        super().__init__(name=name)
        self.tags: list[int] = []
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]
        

class GeoSurface(GeoObject):
    '''GeoVolume is an interface to the GMSH CAD kernel. It does not reprsent Emerge
    specific geometry data.'''
    dim = 2
    _default_name: str = 'GeoSurface'
    
    @property
    def selection(self) -> FaceSelection:
        return FaceSelection(self.tags)
    
    def __init__(self, tag: int | list[int], name: str | None = None):
        super().__init__(name=name)
        self.tags: list[int] = []
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]

class GeoPolygon(GeoSurface):
    _default_name: str = 'GeoPolygon'
    
    def __init__(self,
                 tags: list[int],
                 name: str | None = None):
        super().__init__(tags, name=name)
        self.points: list[int] = []
        self.lines: list[int] = []


############################################################
#                      SHORT FUNCTIONS                     #
############################################################

def select(*items: Selection | GeoObject) -> Selection:
    """Generate a selection from a series of selections and/or objects that share the same dimension.

    Raises:
        ValueError: Raised if the dimensions provided are not consistent

    Returns:
        Selection: An output selection object.
    """
    dim = items[0].dim
    tags = []
    for item in items:
        if item.dim!=dim:
            raise ValueError(f'Cannot group of objects with a dissimilar dimensions. Trying to include {item} in a list of dimension {dim}.')
        tags.extend(item.tags)
    return Selection.from_dim_tags(dim, tags)
            


"""A Python based FEM solver.
Copyright (C) 2025 Robert Fennis

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, see
<https://www.gnu.org/licenses/>.

"""
############################################################
#                    WARNING SUPPRESSION                   #
############################################################

import warnings

warnings.filterwarnings(
    "ignore",
    category=DeprecationWarning,
    message="builtin type swigvarlink.*"
)

############################################################
#               HANDLE ENVIRONMENT VARIABLES              #
############################################################

import os

__version__ = "2.0.0"

NTHREADS = "1"
os.environ["EMERGE_STD_LOGLEVEL"] = os.getenv("EMERGE_STD_LOGLEVEL", default="INFO")
os.environ["EMERGE_FILE_LOGLEVEL"] = os.getenv("EMERGE_FILE_LOGLEVEL", default="DEBUG")
os.environ["OMP_NUM_THREADS"] = os.getenv("OMP_NUM_THREADS", default="1")
os.environ["MKL_NUM_THREADS"] = os.getenv("MKL_NUM_THREADS", default="4")
os.environ["OPENBLAS_NUM_THREADS"] = NTHREADS
os.environ["VECLIB_NUM_THREADS"] = NTHREADS
os.environ["VECLIB_MAXIMUM_THREADS"] = NTHREADS
os.environ["NUMEXPR_NUM_THREADS"] = NTHREADS
os.environ["NUMBA_NUM_THREADS"] = os.getenv("NUMBA_NUM_THREADS", default="4")
os.environ.setdefault("NUMBA_THREADING_LAYER", "workqueue")


############################################################
#                      IMPORT MODULES                     #
############################################################

from ._emerge.logsettings import LOG_CONTROLLER
from loguru import logger

LOG_CONTROLLER.set_default()
logger.debug('Importing modules')
LOG_CONTROLLER._set_log_buffer()

import gmsh
from ._emerge.simmodel import Simulation, SimulationBeta
#from ._emerge.material import Material, FreqCoordDependent, FreqDependent, CoordDependent
from ._emerge import bc
from ._emerge.solver import SolverBicgstab, SolverGMRES, SolveRoutine, ReverseCuthillMckee, Sorter, SolverPardiso, SolverUMFPACK, SolverSuperLU, EMSolver
from ._emerge.cs import CoordinateSystem, CS, GCS, Plane, Axis, XAX, YAX, ZAX, XYPLANE, XZPLANE, YZPLANE, YXPLANE, ZXPLANE, ZYPLANE, cs
from ._emerge.coord import Line
from ._emerge import geo
from ._emerge.selection import Selection, FaceSelection, DomainSelection, EdgeSelection
from ._emerge.geometry import select
from ._emerge.mth.common_functions import norm, coax_rout, coax_rin, dot, cross
from ._emerge.periodic import RectCell, HexCell
from ._emerge.mesher import Algorithm2D, Algorithm3D
#from . import lib
from ._emerge.howto import _HowtoClass
from ._emerge.emerge_update import update_emerge
from .auxilliary.touchstone import TouchstoneData
from emsutil import isola, rogers, const, lib
from emsutil.material import Material, MatProperty, FreqDependent, CoordDependent, FreqCoordDependent
import emsutil.plot as plot
from emsutil import EMergeTheme
from emsutil import themes

howto = _HowtoClass()

logger.debug('Importing complete!')

from ._emerge.install_check import run_installation_checks

run_installation_checks()
############################################################
#                         CONSTANTS                        #
############################################################

CENTER = geo.Alignment.CENTER
CORNER = geo.Alignment.CORNER
EISO = lib.EISO
EOMNI = lib.EOMNI


import numpy as np
from .optimized import compute_distances
from itertools import groupby

def signed_area(x: np.ndarray, y: np.ndarray) -> float:
    """Compute the signed area of a polygon defined by vertices (x, y)."""
    return 0.5 * np.sum(x * np.roll(y, -1) - y * np.roll(x, -1))

class Loop:
    """This class 
    """
    def __init__(self, xs: np.ndarray, ys: np.ndarray):
        self.x: np.ndarray = xs
        self.y: np.ndarray = ys
        if self.x[0]==self.x[-1]:
            self.x = self.x[:-1]
            self.y = self.y[:-1]
        self.N: int = self.x.shape[0]
        self.solid: bool = False

    def cleanup(self) -> None:
        """
        Remove consecutive duplicate points where both x[i+1] == x[i] and y[i+1] == y[i].
        Updates self.x, self.y, and self.N.
        """
        if self.N == 0:
            return

        mask = np.ones(self.N, dtype=bool)
        mask[1:] = ~((self.x[1:] == self.x[:-1]) & (self.y[1:] == self.y[:-1]))

        self.x = self.x[mask]
        self.y = self.y[mask]
        self.N = self.x.shape[0]
    
    def split(self, merge_margin: float = 1e-9) -> tuple[list[tuple[list[float],list[float]]], list[tuple[list[float],list[float]]]]:
        """Splits the loop into an Add and Remove list of (x,y) points.

        Args:
            merge_margin (float, optional): The distance below which points are considered the same. Defaults to 1e-9.

        Returns:
            tuple[list[tuple[list[float],list[float]]], list[tuple[list[float],list[float]]]]: _description_
        """
        self.cleanup()
        self.unique = np.ones_like(self.x, dtype=np.int32)
        
        ds = compute_distances(self.x, self.y, 0*self.x)
        
        conn = dict()
        # For each polygon node index
        for i in range(self.N):
            # get a list of each connected node that is not itself.
            dsv = np.concatenate((ds[i,:i],[1.0,],ds[i,i+1:]))
            # Truth map of those that are the same
            connected = dsv < merge_margin
            # Map node id -> list of all nodes that are the same [j, k, l]
            conn[i] = list(np.argwhere(connected).flatten())
            # If at least one node is the same, set all unique flags to 0
            if len(conn[i])>0:
                self.unique[np.argwhere(connected)] = 0
        
        # if there are no similar nodes, just return the loop
        if sum(self.unique)==self.unique.shape[0]:
            return ([(self.x, self.y),], [])

        # Extract loop sections
        loop_sections = []
        for unique, id_iter in groupby(np.arange(self.N), key = lambda x: self.unique[x]):
            ids = list(id_iter)
            if unique==0:
                continue
            # get all non-overlapping edge sections [1,2,3,4],[8,9,10] etc.
            loop_sections.append([int(i) for i in ids])
        
        # construct a connectivity dictionary
        loops = dict()
        keys = []
        # For each non-overlapping polygon loop section
        for segment_ids in loop_sections:
            # Get the first next segment and find which nodes are the same as it
            conn_end = [int(i+1) for i in conn.get(segment_ids[-1]+1, [])]
            # Then store the next node ast the connected segment id + 1
            loops[segment_ids[0]] = (segment_ids + [(segment_ids[-1]+1)%self.N,], conn_end)
            keys.append(segment_ids[0])
        
        loops_out = []
        while keys:
            startid = keys.pop(0)
            loop, ce = loops[startid]
            
            new_id = sorted(ce)[0]
            # stitching
            while True:
                if new_id==loop[0]:
                    loops_out.append(loop)
                    break
                
                connect_to, cce = loops[new_id]
                keys.remove(new_id)
                loop.extend(connect_to)
                if len(cce)==0:
                    cce = [0,]
                
                new_id = sorted(cce)[0]
            
        areas = [signed_area(self.x[loop], self.y[loop]) for loop in loops_out]
        np_areas = np.array(areas)
        abs_np_areas = np.abs(np_areas)
        
        add_loops = []
        remove_loops = []
        
        sign_biggest = np.sign(np_areas[np.argwhere(abs_np_areas==np.max(abs_np_areas))[0]])
        
        for (loop, A) in zip(loops_out, areas):
            
            if np.sign(A)*sign_biggest > 0:
                add_loops.append(loop)
            else:
                remove_loops.append(loop)
        
        output_add = [(self.x[loop], self.y[loop]) for loop in add_loops]
        output_remove = [(self.x[loop], self.y[loop]) for loop in remove_loops]
        
        return output_add, output_remove


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.
from __future__ import annotations
from .mesh3d import Mesh3D
from .geometry import GeoObject, _GeometryManager, _GEOMANAGER
from .dataset import SimulationDataset
from loguru import logger
from typing import Any
import numpy as np
from .selection import _CALC_INTERFACE
import gmsh

class _SimStateCollection:
    
    def __init__(self):
        self.states: list[SimState] = []
        self.active: SimState | None = None

    def sign_on(self, state: SimState) -> None:
        if state not in self.states:
            self.states.append(state)
        self.active = state
            
_GLOBAL_SIMSTATES = _SimStateCollection()

class SimState:
    
    def __init__(self, modelname: str):
        self.modelname: str = modelname
        self.mesh: Mesh3D = Mesh3D()
        self.geos: list[GeoObject] = []
        self.data: SimulationDataset = SimulationDataset()
        self.params: dict[str, float] = dict()
        self._stashed: SimulationDataset | None = None
        self.manager: _GeometryManager = _GEOMANAGER
        self.sign_on()
    
    def sign_on(self):
        _GLOBAL_SIMSTATES.sign_on(self)
        _CALC_INTERFACE._ifobj = self

    @property
    def current_geo_state(self) -> list[GeoObject]:
        return self.manager.all_geometries()
    
    def reset_geostate(self) -> None:
        _GEOMANAGER.reset(self.modelname)
        self.clear_mesh()
        
    def init(self) -> None:
        """Initializes the Simstate to a clean starting point.
        """
        self.mesh = Mesh3D()
        self.geos = []
        self.reset_geostate()
        self.init_data()
        self.sign_on()
        
    def stash(self) -> None:
        """Stashes the simstate data to run simulations and restore it later.
        """
        self._stashed = self.data
        self.data = SimulationDataset()
    
    def set_parameters(self, parameters: dict[str, float]) -> None:
        """ Define the simulation parameter sweep to a set of outer variables
        defined by a dictionary of string, float pairs.

        Args:
            parameters (dict[str, float]): The parameter sweep slice.
        """
        self.params = parameters
        
    def init_data(self) -> None:
        """Initializes a the dataset with the current parameters
        """
        self.data.sim.new(**self.params)
        
    def reload(self) -> SimulationDataset:
        """Reload stashed data into the simstate memory

        Returns:
            SimulationDataset: _description_
        """
        old = self._stashed
        self.data = self._stashed
        self._stashed = None
        return old
    
    def reset_mesh(self) -> None:
        """Resets and clears the current mesh.
        """
        self.mesh = Mesh3D()
        
    def set_mesh(self, mesh: Mesh3D) -> None:
        """ Overwrite the mesh with a new one."""
        self.mesh = mesh
        
    def set_geos(self, geos: list[GeoObject]) -> None:
        """Activate a set of geometry objects to the simstate geometry field.

        Args:
            geos (list[GeoObject]): _description_
        """
        self.geos = geos
        _GEOMANAGER.set_geometries(geos)

    def clear_mesh(self) -> None:
        """resets the current mesh object to an empty one.
        """
        self.mesh = Mesh3D()
        
    def store_geometry_data(self) -> None:
        """Saves the current geometry state to the simulatin dataset
        """
        logger.trace('Storing geometries in data.sim')
        self.geos = self.current_geo_state
        self.data.sim['geos'] = self.geos
        self.data.sim['mesh'] = self.mesh
        
    def get_dataset(self) -> dict[str, Any]:
        """Create a dict of the file data to store to the harddrive.

        Returns:
            dict[str, Any]: _description_
        """
        return dict(simdata=self.data, mesh=self.mesh)
    
    def load_dataset(self, dataset: dict[str, Any]):
        """Load the data from a dataset

        Args:
            dataset (dict[str, Any]): _description_
        """
        self.data = dataset['simdata']
        self.mesh = dataset['mesh']
        
    def activate(self, _indx: int | None = None, **variables):
        """Searches for the permutaions of parameter sweep variables and sets the current geometry to the provided set."""
        if _indx is not None:
            dataset = self.data.sim.index(_indx)
        else:
            dataset = self.data.sim.find(**variables)
        
        variables = ', '.join([f'{key}={value}' for key,value in dataset.vars.items()])
        logger.info(f'Activated entry with variables: {variables}')
        self.set_mesh(dataset['mesh'])
        self.set_geos(dataset['geos'])
        
        return self
    

    ############################################################
    #                       GMSH LIKE METHODS                  #
    ############################################################

    def getCenterOfMass(self, dim: int, tag: int) -> np.ndarray:
        if self.mesh.defined is False:
            return gmsh.model.occ.getCenterOfMass(dim, tag)
        return self.mesh.dimtag_to_center[(dim, tag)]
    
    def getPoints(self, dimtags: list[tuple[int, int]]) -> list[np.ndarray]:
        """Returns a

        Args:
            dimtags (list[tuple[int, int]]): _description_

        Returns:
            list[np.ndarray]: _description_
        """
        points = []
        id_set = []
        for dt in dimtags:
            id_set.append(self.mesh.dimtag_to_nodes[dt])
        ids = np.unique(np.concatenate(id_set))
        points = [self.mesh.nodes[:,i] for i in ids]
        return points
    
    def getBoundingBox(self, dim: int, tag: int) -> tuple[float, float, float, float, float, float]:
        """Returns the bounding box corresponding to entity defined by (dim,tag)

        Args:
            dim (int): _description_
            tag (int): _description_

        Returns:
            tuple[float, float, float, float, float, float]: _description_
        """
        if self.mesh.defined is False:
            return gmsh.model.occ.getBoundingBox(dim, tag)
        return self.mesh.dimtag_to_bb[(dim, tag)]
    
    def getNormal(self, facetag: int) -> np.ndarray:
        """Returns the normal vector of a facetag

        Args:
            facetag (int): _description_

        Returns:
            np.ndarray: _description_
        """
        if self.mesh.defined:
            return self.mesh.ftag_to_normal[facetag]
        else:
            return np.array(gmsh.model.getNormal(facetag, (0,0)))
    
    def getCharPoint(self, facetag: int) -> np.ndarray:
        """Returns a coordinate that is always on the surface and roughly in the center

        Args:
            facetag (int): _description_

        Returns:
            np.ndarray: _description_
        """
        return self.mesh.ftag_to_point[facetag]
    
    def getArea(self, tag: int) -> float:
        if self.mesh.defined is True:
            area = sum([self.mesh.areas[tri] for tri in self.mesh.ftag_to_tri[tag]])
            return area
        else:
            area = gmsh.model.occ.getMass(2,tag)
            return area


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np

def norm(Field: np.ndarray) -> np.ndarray:
    """ Computes the complex norm of a field (3,N)

    Args:
        Field (np.ndarray): The input field, shape (3,N)

    Returns:
        np.ndarray: The complex norm in shape (N,)
    """
    return np.sqrt(np.abs(Field[0,:])**2 + np.abs(Field[1,:])**2 + np.abs(Field[2,:])**2)

def coax_rout(rin: float,
              eps_r: float = 1,
              Z0: float = 50) -> float:
    """Computes the outer radius given a dielectric constant, inner radius and characteristic impedance

    Args:
        rin (float): The inner radius
        eps_r (float, optional): The dielectric permittivity. Defaults to 1.
        Z0 (float, optional): The impedance. Defaults to 50.

    Returns:
        float: The outer radius
    """
    return rin*10**(Z0*np.sqrt(eps_r)/138)

def coax_rin(rout: float,
              eps_r: float = 1,
              Z0: float = 50) -> float:
    """Computes the inner radius given a dielectric constant, outer radius and characteristic impedance

    Args:
        rin (float): The outer radius
        eps_r (float, optional): The dielectric permittivity. Defaults to 1.
        Z0 (float, optional): The impedance. Defaults to 50.

    Returns:
        float: The inner radius
    """
    return rout/10**(Z0*np.sqrt(eps_r)/138)

def _move_3_to_last(x: np.ndarray) -> np.ndarray:
    """
    Ensure the axis of length 3 is the last axis.
    Accepts shapes (3,), (3, N), (N, 3).
    """
    x = np.asarray(x)

    if x.ndim == 1:
        if x.shape[0] != 3:
            raise ValueError(f"1D input must have shape (3,), got {x.shape}")
        return x  # shape (3,)

    if x.ndim != 2:
        raise ValueError(f"Input must be 1D or 2D, got {x.ndim}D with shape {x.shape}")

    axes_3 = [i for i, s in enumerate(x.shape) if s == 3]
    if len(axes_3) != 1:
        raise ValueError(f"Input must have exactly one axis of length 3, got shape {x.shape}")

    axis_3 = axes_3[0]
    if axis_3 == x.ndim - 1:
        return x  # already (..., 3)
    else:
        # swap the 3-axis to the last position: (3, N) -> (N, 3)
        return np.swapaxes(x, axis_3, x.ndim - 1)


def dot(A, B):
    """
    Dot product along the axis of length 3.

    A, B can be:
      - (3,)        : single 3-vector
      - (3, N)      : 3-by-N, treated as N vectors of length 3
      - (N, 3)      : N-by-3, treated as N vectors of length 3

    Returns:
      - scalar for (3,)·(3,)
      - (N,) for pairwise dot of N vectors.
    """
    A = _move_3_to_last(np.asarray(A))
    B = _move_3_to_last(np.asarray(B))

    # Broadcast over non-3 axes (e.g. (N,3) with (3,) -> (N,3))
    A, B = np.broadcast_arrays(A, B)

    return np.sum(A * B, axis=-1)


def cross(A, B):
    """
    Cross product along the axis of length 3.

    A, B can be:
      - (3,)
      - (3, N)
      - (N, 3)

    Returns:
      - (3,) for single cross product
      - (3, N) for pairwise cross of N vectors.
    """
    A = _move_3_to_last(np.asarray(A))
    B = _move_3_to_last(np.asarray(B))

    # Broadcast over non-3 axes
    A, B = np.broadcast_arrays(A, B)

    a1, a2, a3 = A[..., 0], A[..., 1], A[..., 2]
    b1, b2, b3 = B[..., 0], B[..., 1], B[..., 2]

    c1 = a2 * b3 - a3 * b2
    c2 = a3 * b1 - a1 * b3
    c3 = a1 * b2 - a2 * b1

    return np.array((c1, c2, c3))


""" DIFFERENTIAL AND COMMON MODE SIMULATIONS

In this simulation we look at a trivial model with two parallel microstripline traces. We show how we can
combine two ports into one to yield common and differential mode ports.

"""

import emerge as em
from emerge.plot import plot_sp

mm = 0.001  # meter units helper: 1 mm

model = em.Simulation('DiffCommon')  # simulation container
model.check_version('2.0.0') # Checks version compatibility.

# PCB stack: 1 mm substrate, RO3003 dielectric, 0.05 mm copper thickness, thick traces enabled
pcb = em.geo.PCB(1, mm, material=em.lib.DIEL_RO3003, trace_thickness=0.05*mm, thick_traces=True)

w0 = pcb.calc.z0(50)  # compute trace width for 50 Ω single-ended impedance

# Route two parallel single-ended microstrips. Store endpoints to attach lumped ports.
# Upper line: p1 -> straight -> p3
pcb.new(0, 2, w0, (1,0)).store('p1').straight(20).store('p3')
# Lower line: p2 -> straight -> p4
pcb.new(0, -2, w0, (1,0)).store('p2').straight(20).store('p4')

traces = pcb.compile_paths(merge=True)  # polygonize traces for meshing, merged for convenience

# Create conservative board boundary and air box margins
pcb.determine_bounds(5, 5, 5, 5)

diel = pcb.generate_pcb()    # substrate solid
air = pcb.generate_air(5)    # surrounding air region

# Create four lumped-port terminals at the stored endpoints
lp1 = pcb.lumped_port(pcb.load('p1'))
lp2 = pcb.lumped_port(pcb.load('p2'))
lp3 = pcb.lumped_port(pcb.load('p3'))
lp4 = pcb.lumped_port(pcb.load('p4'))

model.commit_geometry()  # hand geometry to the solver
model.mw.set_frequency_range(2e9, 4e9, 11)  # coarse sweep: 2–4 GHz, 5 points

# Improve mesh around conductor edges
model.mesher.set_boundary_size(traces, 1*mm)

model.generate_mesh()  # build volume mesh

model.view()  # optional mesh preview

# Attach lumped ports to solver boundary conditions, numbered 1..4
model.mw.bc.LumpedPort(lp1, 1)
model.mw.bc.LumpedPort(lp2, 2)
model.mw.bc.LumpedPort(lp3, 3)
model.mw.bc.LumpedPort(lp4, 4)

data = model.mw.run_sweep(True, 4)  # frequency sweep, parallel workers=4

g = data.scalar.grid  # reshape scalar results to an N-D grid helper

# Plot raw single-ended S-parameters before any port combination
plot_sp(g.freq, [g.S(1,1), g.S(2,1), g.S(3,1), g.S(4,1)], labels=['S11','S21','S31','S41'])

# Combine port pairs (1,2) and (3,4).
# After this, port indices are re-interpreted as modal ports:
#   1: differential of original (1,2)
#   2: common of original (1,2)
#   3: differential of original (3,4)
#   4: common of original (3,4)
g.combine_ports(1,2)
g.combine_ports(3,4)

# Plot S-parameters of the new modal ports. Useful to see DM/CM isolation and balance.
plot_sp(g.freq, [g.S(1,1), g.S(2,1), g.S(3,1), g.S(4,1)], labels=['S11','S21','S31','S41'])

field = data.field.find(freq=3e9)  # pick a field solution at 3 GHz

# Visualize single-ended excitations first: excite each original port, show Ez cutplane
for i, text in enumerate(['Port 1','Port 2','Port 3','Port 4']):
    field.excite_port(i+1)  # single-ended excitation i+1
    model.display.add_objects(*model.all_geos())
    model.display.animate().add_field(field.cutplane(1*mm, z=-0.7*mm).scalar('Ez','real'), symmetrize=True)
    model.display.add_title(text)
    model.display.show()

# Combine field ports into modal pairs as above.
# After this, field.excite_port(1) gives differential of (1,2), 2 gives common of (1,2), etc.
field.combine_ports(1,2)
field.combine_ports(3,4)

# Visualize modal excitations: DM and CM field patterns for both ends
for i, text in enumerate(['Port 1 differential mode','Port 1 common mode','Port 2 differential mode','Port 2 common mode']):
    field.excite_port(i+1)  # modal excitation i+1 after combination
    model.display.add_objects(*model.all_geos())
    model.display.animate().add_field(field.cutplane(1*mm, z=-0.7*mm).scalar('Ez','real'), symmetrize=True)
    model.display.add_title(text)
    model.display.show()


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from loguru import logger
import sys
from typing import Literal, Generator
from pathlib import Path
import os
from collections import deque
import platform
import importlib
import multiprocessing

packages = ["numba", "numpy", "scipy", "gmsh", "joblib","matplotlib","pyvista","mkl","cloudpickle","scikit-umfpack","nvidia-cudss-cu12","nvmath-python[cu12]","cupy-cuda12x","ezdxf"]

def get_version(pkg_name):
    try:
        module = importlib.import_module(pkg_name)
        return getattr(module, "__version__", "unknown")
    except ImportError:
        return "not installed"
    
_LOG_BUFFER = deque()

def _log_sink(message):
    _LOG_BUFFER.append(message)
    

############################################################
#                          FORMATS                         #
############################################################

TRACE_FORMAT = (
    "{time:ddd YY/MM/DD HH:mm:ss.SSSS} {level:<7} {thread.id:<15} {line:>4}: "
    "{message}"
)

DEBUG_FORMAT = (
    "<green>{elapsed}</green>  <level>{level:<7}</level>: "
    "<level>{message}</level>"
)
INFO_FORMAT = (
    "<green>{elapsed}</green>  <level>{level:<7}</level>: "
    "<level>{message}</level>"
)
WARNING_FORMAT = (
    "<green>{elapsed}</green>  <level>{level:<7}</level>: "
    "<level>{message}</level>"
)
ERROR_FORMAT = (
    "<green>{elapsed}</green>  <level>{level:<7}</level>: "
    "<level>{message}</level>"
)
FORMAT_DICT = {
    'TRACE': TRACE_FORMAT,
    'DEBUG': DEBUG_FORMAT,
    'INFO': INFO_FORMAT,
    'WARNING': WARNING_FORMAT,
    'ERROR': ERROR_FORMAT,
}

LLTYPE = Literal['TRACE','DEBUG','INFO','WARNING','ERROR'] 


############################################################
#                      LOG CONTROLLER                     #
############################################################

class LogController:
    
    def __init__(self):
        logger.remove()
        self.std_handlers: list[int] = []
        self.file_handlers: list[int] = []
        self.level: str = 'INFO'
        self.file_level: str = 'INFO'
    
    def set_default(self):
        value = os.getenv("EMERGE_STD_LOGLEVEL", default="INFO")
        self.set_std_loglevel(value)

    def add_std_logger(self, loglevel: LLTYPE) -> None:
        handle_id = logger.add(sys.stderr, 
                level=loglevel, 
                format=FORMAT_DICT.get(loglevel, INFO_FORMAT))
        self.std_handlers.append(handle_id)

    def _set_log_buffer(self):
        logger.add(_log_sink)
        
    def _flush_log_buffer(self):
        for msg in list(_LOG_BUFFER):
            logger.opt(depth=6).log(msg.record["level"].name, msg.record["message"])
        _LOG_BUFFER.clear()
        
    def _sys_info(self) -> None:
        for pkg in packages:
            logger.trace(f" (!) {pkg} version: {get_version(pkg)}")

        logger.trace(f" (!) OS: {platform.system()} {platform.release()} ({platform.version()})")
        logger.trace(f" (!) Architecture: {platform.machine()} ({platform.architecture()[0]})")
        logger.trace(f" (!) Processor: {platform.processor()}")
        logger.trace(f" (!) Python build: {platform.python_build()}")
        logger.trace(f" (!) Python version: {platform.python_version()} [{sys.version}]")
        logger.trace(f" (!) CPU cores: {multiprocessing.cpu_count()}")
        
    def set_std_loglevel(self, loglevel: str):
        handler = {"sink": sys.stdout, 
                   "level": loglevel, 
                   "format": FORMAT_DICT.get(loglevel, INFO_FORMAT)}
        logger.configure(handlers=[handler]) # type: ignore
        self.level = loglevel
        os.environ["EMERGE_STD_LOGLEVEL"] = loglevel

    def set_write_file(self, path: Path, loglevel: str = 'TRACE'):
        handler_id = logger.add(str(path / 'logging.log'), mode='w', level=loglevel, format=FORMAT_DICT.get(loglevel, INFO_FORMAT), colorize=False, backtrace=True, diagnose=True)
        self.file_handlers.append(handler_id)
        self.file_level = loglevel
        os.environ["EMERGE_FILE_LOGLEVEL"] = loglevel

class DebugCollector:
    """The DebugController is used by EMerge to collect heuristic
    warnings for detections of things that might be causing problems but aren't
    guaranteed to cause them. These logs will be printed at the end of a simulation
    to ensure that users are aware of them if they abort simulations.
    
    """
    def __init__(self):
        self.reports: list[str] = []
    
    @property
    def any_warnings(self) -> bool:
        return len(self.reports)>0
    
    def add_report(self, message: str):
        self.reports.append(message)
        
    def all_reports(self) -> Generator[tuple[int, str], None, None]:
        
        for i, message in enumerate(self.reports):
            yield i+1, message


############################################################
#                        SINGLETONS                       #
############################################################


LOG_CONTROLLER = LogController()
DEBUG_COLLECTOR = DebugCollector()


import gmsh
from ..geometry import GeoPoint, GeoEdge, GeoVolume, GeoSurface, GeoObject
from ..selection import FaceSelection
from .shapes import Box
from .operations import unite

from pathlib import Path
import numpy as np
from typing import Callable

def _select_num(num1: float | None, num2: float | None) -> float:
    if num1 is None and num2 is None:
        return 0.0
    if isinstance(num1, float) and num2 is None:
        return num1
    if num1 is None and isinstance(num2, float):
        return num2
    return max(num1, num2)

class _FaceSliceSelector:
    
    def __init__(self, face_numbers: list[int], selector: Callable):
        self.numbers: list[str] = face_numbers
        self.selector: Callable = selector
        
    def __getitem__(self, slice) -> FaceSelection:
        nums = self.numbers.__getitem__(slice)
        if isinstance(nums, int):
            nums = [nums,]
        return self.selector(*nums)
    
class StepVolume(GeoVolume):
    """The StepVoume class extens the EMerge GeoVolume class to add easier
    face selection functionalities based on numbers as face names are not 
    imported currently

    Args:
        GeoVolume (_type_): _description_

    Returns:
        _type_: _description_
    """
    
        
    @property
    def _face_numbers(self) -> list[int]:
        return sorted([int(name[4:]) for name in self._face_pointers.keys()])
    
    @property
    def face_slice(self) -> _FaceSliceSelector:
        return _FaceSliceSelector(self._face_numbers, self.faces)
    
    def faces(self, *numbers: int) -> FaceSelection:
        """Select a set of faces by number

        Returns:
            FaceSelection: _description_
        """
        names = [f'Face{num}' for num in numbers]
        return super().faces(*names)
        

class STEPItems:
    """STEPItems imports geometries form a STEP file and exposes them to the user.
    
    """
    def __init__(self, name: str, filename: str, unit: float = 1.0):
        """Imports the provided STEP file.
        Specify the unit in case of scaling issues where mm units are not taken into consideration.

        Args:
            filename (str): The filename
            unit (float, optional): The STEP file size unit. Defaults to 1.0.

        Raises:
            FileNotFoundError: If a file does not exist
        """
        self.name: str = name

        stl_path = Path(filename)
        gmsh.option.setNumber("Geometry.OCCScaling", unit)
        gmsh.option.setNumber("Geometry.OCCImportLabels", 1) 

        if not stl_path.exists:
            raise FileNotFoundError(f'File with name {stl_path} does not exist.')
        
        dimtags = gmsh.model.occ.import_shapes(filename, format='step')
        
        self.points: list[GeoPoint] = []
        self.edges: list[GeoEdge] = []
        self.surfaces: list[GeoSurface] = []
        self.volumes: list[GeoVolume] = []
        
        i = 0
        for dim, tag in dimtags:
            name = gmsh.model.getPhysicalName(dim, tag) #for now, this doesn't actually ever work.
            if name == '':
                name = f'Obj{i}'
                i+=1
            if dim == 0:
                self.points.append(GeoPoint(tag, name=f'{self.name}_{name}'))
            elif dim == 1:
                self.edges.append(GeoEdge(tag, name=f'{self.name}_{name}'))
            elif dim == 2:
                self.surfaces.append(GeoSurface(tag, name=f'{self.name}_{name}'))
            elif dim == 3:
                self.volumes.append(StepVolume(tag, name=f'{self.name}_{name}'))
        
        gmsh.model.occ.synchronize()
        
    @property
    def dictionary(self) -> dict[str, GeoObject]:
        return {obj.name: obj for obj in self.objects}
    
    @property
    def objects(self) -> tuple[GeoObject,...]:
        """Returns a list of all objects in the STEP file

        Returns:
            tuple[GeoObject,...]: _description_
        """
        return tuple(self.points+self.edges+self.surfaces+self.volumes)
    
    def __getitem__(self, name: str) -> GeoObject | None:
        return self.dictionary.get(name, None)
    
    def as_volume(self) -> StepVolume:
        """Returns the 3D volumetric part of the STEP file as a single geometry

        Returns:
            StepVolume: The resultant StepVolume(GeoVolume) object.
        """
        if len(self.volumes)==1:
            return self.volumes[0]
        return unite(*self.volumes)._auto_face_tag()
    
    def as_surface(self) -> GeoSurface:
        """Returns the 2D surface part of the STEP file as a single geometry


        Returns:
            GeoSurface: The resultant GeoSurface object
        """
        if len(self.surfaces)==1:
            return self.surfaces[0]
        return unite(*self.surfaces)
    
    def as_edge(self) -> GeoEdge:
        """Returns the 1D Edge part of the STEP file as a single geometry

        Returns:
            GeoEdge: The resultant GeoEdge object
        """
        if len(self.edges)==1:
            return self.edges[1]
        return unite(*self.edges)
    
    def as_point(self) -> GeoPoint:
        """Returns the 0D Point part of the STEP file as a single geometry

        Returns:
            GeoPoint: The resultant GeoPoint object.
        """
        if len(self.points)==1:
            return self.points[0]
        return unite(*self.points)
    
    def enclose_params(self, 
                margin: float = None,
                x_margins: tuple[float, float] = (None, None),
                y_margins: tuple[float, float] = (None, None),
                z_margins: tuple[float, float] = (None, None)) -> tuple[float,float,float,tuple[float,float,float]]:
        """Create an enclosing bounding box for the step model.

        Args:
            margin (float, optional): _description_. Defaults to 0.
            x_margins (tuple[float, float], optional): _description_. Defaults to (0., 0.).
            y_margins (tuple[float, float], optional): _description_. Defaults to (0., 0.).
            z_margins (tuple[float, float], optional): _description_. Defaults to (0., 0.).

        Returns:
            Box: _description_
        """
        xminm = _select_num(margin, x_margins[0])
        xmaxm = _select_num(margin, x_margins[1])
        yminm = _select_num(margin, y_margins[0])
        ymaxm = _select_num(margin, y_margins[1])
        zminm = _select_num(margin, z_margins[0])
        zmaxm = _select_num(margin, z_margins[1])
        
        xmin = 1000000
        xmax = -1000000
        ymin = 1000000
        ymax = -1000000
        zmin = 1000000
        zmax = -1000000
        
        for obj in self.objects:
            for dim, tag in obj.dimtags:
                x1, y1, z1, x2, y2, z2 = gmsh.model.occ.getBoundingBox(dim, tag)
                xmin = min(xmin, x1)
                xmax = max(xmax, x2)
                ymin = min(ymin, y1)
                ymax = max(ymax, y2)
                zmin = min(zmin, z1)
                zmax = max(zmax, z2)
        
        width = xmax-xmin + xminm + xmaxm
        depth = ymax-ymin + yminm + ymaxm
        height = zmax -zmin + zminm + zmaxm
        return width, depth, height, (xmin-xminm, ymin-yminm, zmin-zminm)
    
    def enclose(self, 
                margin: float = None,
                x_margins: tuple[float, float] = (None, None),
                y_margins: tuple[float, float] = (None, None),
                z_margins: tuple[float, float] = (None, None)) -> Box:
        """Create an enclosing bounding box for the step model.

        Args:
            margin (float, optional): _description_. Defaults to 0.
            x_margins (tuple[float, float], optional): _description_. Defaults to (0., 0.).
            y_margins (tuple[float, float], optional): _description_. Defaults to (0., 0.).
            z_margins (tuple[float, float], optional): _description_. Defaults to (0., 0.).

        Returns:
            Box: _description_
        """
        
        return Box(*self.enclose_params(margin, x_margins, y_margins, z_margins)).background()
    